//
// pvbmp.c - Routines for BMP (and paste-from-clipboard) support in 
// Purpleview.
// This could probably use some cleanup -- there's a lot of code that's
// largely redundant.
//

#include <windows.h>
#include <tchar.h>
#include <stdio.h>

#include "purpleview.h"
#include <strsafe.h>

// Determine if this DIB needs to be converted to our restricted form.
// 
// We define a  "restricted" DIB follows:
// - V3, with 40-byte header
// biBitCount = 1, 4, 8, or 24
// no compression
// no palette if > 8bitsperpixel
// the bits section must immediately follow the palette
//   (or header, if no palette)

// -----------
// bits_offset==0 : offset unknown
//
// returns 1 if okay
// returns 0 if needs fixed
static int is_restricted_dib(LPBITMAPINFOHEADER lpdib, unsigned int dib_size, unsigned int bits_offset_1)
{
	unsigned int bitcount;
	unsigned int clrused;
	unsigned int bits_offset;

	if(dib_size<40) return 0;
	if(lpdib->biSize!=40) return 0;

	bitcount=lpdib->biBitCount;
	if(lpdib->biCompression != BI_RGB) return 0;
	if(bitcount!=1 && bitcount!=4 && bitcount!=8 && bitcount!=24) return 0;

	clrused= lpdib->biClrUsed;
	if(bitcount>=16) {
		if(clrused>0) return 0;
	}
	else {
		if(clrused==0) clrused = 1<<bitcount;
	}
	bits_offset = 40+4*clrused;
	if((bits_offset_1!=0) && (bits_offset!=bits_offset_1)) return 0;
	return 1;
}

// FIXME: This may crash if an invalid RLE file is encountered.
static int rle_uncompress(HWND hwnd,unsigned char* old_lpbits,
						  unsigned char* new_lpbits,
						  int width,int height,int bpp)
{
	int pos; // byte position in rle
	int x,y; // (next) position in new bitmap
	int rowbytes;
	int row_inc;
	unsigned char b1,b2;
	int i;
	unsigned char c;
	int upixels; // uncompressed pixels remaining
	int deltaflag; // set if next pair of pixels is a delta

	rowbytes = pv_get_dib_bytes_per_row(bpp,width);
	row_inc = 1;
	y = 0;
	x=0;
	pos=0;
	upixels=0;
	deltaflag=0;

	while(y>=0 && y<height) {

		b1= old_lpbits[pos];
		b2= old_lpbits[pos+1];
		pos+=2;

		if(deltaflag) { // fixme: this is untested
			x+=b1;
			y+=b2;
			deltaflag=0;
		}
		else if(upixels>0) {
			if(bpp==4) {
				// up to 4 pixels
				for(i=0;i<4;i++) {
					switch(i) {
					case 0: c=(b1&0xf0)>>4; break;
					case 1: c=(b1&0x0f);    break;
					case 2: c=(b2&0xf0)>>4; break;
					case 3: c=(b2&0x0f);
					}
					new_lpbits[y*rowbytes+x/2] |= (x%2)?c:(c<<4);
					x++;
					upixels--;
					if(upixels<=0) break;
				}
			}
			else {
				new_lpbits[y*rowbytes+x]=b1; x++;
				upixels--;
				if(upixels>0) {
					new_lpbits[y*rowbytes+x]=b2; x++;
					upixels--;
				}
			}

		}
		else {
			if(b1==0) { // escape code
				if(b2>=3) {
					upixels=b2;
				}
				else if(b2==0) { // end of line
					x=0;
					y+= row_inc;
				}
				else if(b2==1) { // end of bitmap
					return 1;
				}
				else if(b2==2) { // delta
					deltaflag=1;
				}
			}
			else {   // rle compression
				if(bpp==4) {
					for(i=0;i<b1;i++) {
						c= (i%2)?(b2&0x0f):((b2&0xf0)>>4);
						if(x%2==0) c<<=4;
						new_lpbits[y*rowbytes+x/2] |= c; x++;
					}
				}
				else {
					for(i=0;i<b1;i++) {
						new_lpbits[y*rowbytes+x]=b2; x++;
					}
				}

			}
		}
	}

	return 1;
}


// make a DIB into a restricted DIB
// (We could probably use Windows GDI functions to do this
// much more easily, but this is potentially portable to non-Windows,
// and doesn't rely on Windows doing things correctly, and can
// support DIBs that not all versions of Windows support.)
static int fix_dib(HWND hwnd, struct iinfo_struct *iinfo, struct einfo_struct *einfo)
{
	LPBITMAPINFOHEADER new_lpdib;
	int new_bpp,old_bpp;
	unsigned int new_lpdib_size;
	unsigned int new_bits_size;
	unsigned int new_bits_offs;
	unsigned char* new_lpbits, *old_lpbits;
	RGBQUAD *new_palette;
	DWORD *old_bitfields;

	DWORD bitfield[3];
	unsigned int msb[3];
	unsigned int bitcount[3];
	LONG xdensity, ydensity;

	unsigned char *old_palette;
	int width,height;
	int old_rowbytes;
	int new_rowbytes;
	int compression, has_bitfields;
	int os2;
	int i,j,k,x,c;

	int topdown;
	int palentsize;
	int pal_entries;
	int dib_alpha32;
	unsigned char alphaval;

	dib_alpha32 = 0;

	// figure out how big the new DIB needs to be, and
	// allocate it.

	if(iinfo->lpdib->biSize==12) {
		width= ((BITMAPCOREHEADER*)iinfo->lpdib)->bcWidth;
		height= ((BITMAPCOREHEADER*)iinfo->lpdib)->bcHeight;
		old_bpp= ((BITMAPCOREHEADER*)iinfo->lpdib)->bcBitCount;
		xdensity=0; ydensity=0;
		os2=1;
		compression = BI_RGB;
		palentsize=3;
		if(old_bpp<16) pal_entries = 1<<old_bpp;
		else pal_entries = 0;
	}
	else if(iinfo->lpdib->biSize>=40) {
		width = iinfo->lpdib->biWidth;
		height = iinfo->lpdib->biHeight;
		old_bpp = iinfo->lpdib->biBitCount;
		xdensity=iinfo->lpdib->biXPelsPerMeter;
		ydensity=iinfo->lpdib->biYPelsPerMeter;
		os2=0;
		palentsize=4;
		compression = iinfo->lpdib->biCompression;
		if(iinfo->lpdib->biClrUsed > 0) {
			pal_entries = iinfo->lpdib->biClrUsed;
		}
		else {
			if(old_bpp<16) pal_entries = 1<<old_bpp;
			else pal_entries = 0;
		}

	}
	else {
		pv_mesg(hwnd,MSG_E,_T("Corrupt or unsupported image"));
		return 0;
	}

	switch(compression) {
	case BI_RGB: case BI_RLE4: case BI_RLE8:
		has_bitfields=0;
		break;
	case BI_BITFIELDS:
		has_bitfields=1;
		break;
	default:
		pv_mesg(hwnd,MSG_E,_T("Unsupported compression (%d)"),compression);
		return 0;
	}


	if(height<0) {
		height= -height;
		topdown=1;
	}
	else { topdown=0; }

	switch(old_bpp) {
	case 1: case 4: case 8: case 24:
		new_bpp=old_bpp;
		break;
	case 16: case 32:
		new_bpp=24;
		break;
	default:
		pv_mesg(hwnd,MSG_E,_T("Unsupported bits-per-pixel (%d)"),old_bpp);
		return 0;
	}

	if(old_bpp==32 && einfo->bmp_alpha32 && einfo->use_custom_bg) {
		dib_alpha32=1;
	}

	if(iinfo->lpdib_bits_offs<=0) {
		// we weren't given a pointer to the BITS; need to calculate it
		iinfo->lpdib_bits_offs = iinfo->lpdib->biSize +
			(has_bitfields?12:0) + pal_entries*palentsize;
	}

	// if a 16+ bpp image had a (suggested) palette, we don't need it
	// anymore
	if(new_bpp>=16) pal_entries=0;

	new_rowbytes = pv_get_dib_bytes_per_row(new_bpp,width);
	old_rowbytes = pv_get_dib_bytes_per_row(old_bpp,width);

	new_bits_offs = sizeof(BITMAPINFOHEADER)+ 4*pal_entries;
	new_bits_size = new_rowbytes*height; // unneeded?
	new_lpdib_size = new_bits_offs + new_bits_size;

	new_lpdib = (LPBITMAPINFOHEADER)HeapAlloc(einfo->heap,HEAP_ZERO_MEMORY,new_lpdib_size);
	if(!new_lpdib) {
		pv_mesg(hwnd,MSG_E,_T("Couldn't allocate new bitmap"));
		return 0;
	}

	new_lpdib->biSize = sizeof(BITMAPINFOHEADER);
	new_lpdib->biWidth = width;
	new_lpdib->biHeight = height;
	new_lpdib->biPlanes = 1;
	new_lpdib->biBitCount = new_bpp;
	new_lpdib->biCompression = BI_RGB;
	new_lpdib->biSizeImage = new_bits_size;
	new_lpdib->biXPelsPerMeter = xdensity;
	new_lpdib->biYPelsPerMeter = ydensity;
	new_lpdib->biClrUsed = pal_entries;
	new_lpdib->biClrImportant = 0;


	old_lpbits = &((unsigned char*)iinfo->lpdib)[iinfo->lpdib_bits_offs];
	new_lpbits = &((unsigned char*)new_lpdib)[new_bits_offs];

	old_palette = &((unsigned char*)iinfo->lpdib)[iinfo->lpdib->biSize + (has_bitfields?12:0)];
	new_palette = (RGBQUAD*)&((unsigned char*)new_lpdib)[sizeof(BITMAPINFOHEADER)];

	// copy the palette
	for(i=0; i<pal_entries; i++) {
		if(palentsize==3) {
			new_palette[i].rgbRed = ((RGBTRIPLE*)old_palette)[i].rgbtRed;
			new_palette[i].rgbGreen = ((RGBTRIPLE*)old_palette)[i].rgbtGreen;
			new_palette[i].rgbBlue = ((RGBTRIPLE*)old_palette)[i].rgbtBlue;
		}
		else {
			new_palette[i].rgbRed = ((RGBQUAD*)old_palette)[i].rgbRed;
			new_palette[i].rgbGreen = ((RGBQUAD*)old_palette)[i].rgbGreen;
			new_palette[i].rgbBlue = ((RGBQUAD*)old_palette)[i].rgbBlue;
			new_palette[i].rgbReserved = ((RGBQUAD*)old_palette)[i].rgbReserved;
		}
	}
	

	if(old_bpp==16 || old_bpp==32) {
		if(has_bitfields) {
			old_bitfields = (DWORD*)&((unsigned char*)iinfo->lpdib)[iinfo->lpdib->biSize];
			for(i=0;i<3;i++) bitfield[i]=old_bitfields[i];

			// Analyze the bitfields to find the first ON bit, and
			// the number of ON bits.
			for(k=0;k<3;k++) {  // red, green, blue
				bitcount[k]=0;
				msb[k]=0;
				for(x=0;x<32;x++) {
					if(bitfield[k] & (1<<x)) {
						bitcount[k]++;
						msb[k]=x;
					}
				}
			}
		}
		else if(compression==BI_RGB) {
			// default bitfields
			if(old_bpp==16) {
				bitfield[0]=0x00007c00; bitcount[0]=5; msb[0]=14;
				bitfield[1]=0x000003e0; bitcount[1]=5; msb[1]=9;
				bitfield[2]=0x0000001f; bitcount[2]=5; msb[2]=4;
			}
			else {
				bitfield[0]=0x00ff0000; bitcount[0]=8; msb[0]=23;
				bitfield[1]=0x0000ff00; bitcount[1]=8; msb[1]=15;
				bitfield[2]=0x000000ff; bitcount[2]=8; msb[2]=7;
			}
			has_bitfields=1;
		}
		else {
			pv_mesg(hwnd,MSG_E,_T("Internal: unexpected compression"));
			return 0;
		}
	}


	if(compression==BI_RLE4 || compression==BI_RLE8) {
		rle_uncompress(hwnd,old_lpbits,new_lpbits,width,height,old_bpp);
		goto copydone;
	}


	for(j=0;j<height;j++) {
		if(topdown) k=height-1-j;
		else k=j;
		if(has_bitfields) {
			for(i=0;i<width;i++) {
				unsigned char new_color[3];
				DWORD old_color;

				if(old_bpp==16) {
					old_color =
						(old_lpbits[old_rowbytes*j+2*i  ]   ) |
						(old_lpbits[old_rowbytes*j+2*i+1]<<8);
				}
				else { // 32
					alphaval = old_lpbits[old_rowbytes*j+4*i+3];
					old_color =
						(old_lpbits[old_rowbytes*j+4*i  ]    ) |
						(old_lpbits[old_rowbytes*j+4*i+1]<< 8) |
						(old_lpbits[old_rowbytes*j+4*i+2]<<16) |
						(alphaval<<24);
				}

				for(c=0;c<3;c++) {
					if(msb[c]>7)
						new_color[c] = (unsigned char)((old_color & bitfield[c])>>(msb[c]-7));
					else
						new_color[c] = (unsigned char)((old_color & bitfield[c])<<(7-msb[c]));

					// Copy the bits again for better color accuracy. This will
					// be optimal provided that each color sample has at least 4 bits.
					if(msb[c]+bitcount[c]>7) // pretty much always true
						new_color[c] |= (old_color & bitfield[c])>>( msb[c]+bitcount[c]-7);
					else
						new_color[c] |= (old_color & bitfield[c])<<(7-msb[c]-bitcount[c]);
				}

				if(dib_alpha32) {
					double alpha_flt;
					unsigned char bkcol;

					alpha_flt = ((double)alphaval)/255.0;
					for(x=0;x<3;x++) {
						switch(x) {
						case 0:  bkcol=GetRValue(einfo->custom_bg); break;
						case 1:  bkcol=GetGValue(einfo->custom_bg); break;
						default: bkcol=GetBValue(einfo->custom_bg);
						}

						if(alphaval==0) {
							new_color[x] = bkcol;
						}
						else if(alphaval==255) {
							;   // new_color[x] = new_color[x];
						}
						else {  // do alphamix
							new_color[x] = (int) ( 0.5+
							     (double)bkcol*(1.0-alpha_flt) +
							     (double)new_color[x]*(alpha_flt) );
						}
					}
				}

				new_lpbits[new_rowbytes*k+3*i  ]=new_color[2];
				new_lpbits[new_rowbytes*k+3*i+1]=new_color[1];
				new_lpbits[new_rowbytes*k+3*i+2]=new_color[0];
			}
		}
		else {
			CopyMemory(&new_lpbits[new_rowbytes*k],&old_lpbits[old_rowbytes*j],old_rowbytes);
		}
	}

copydone:

	// final cleanup: modify iinfo

	HeapFree(einfo->heap,0,(LPVOID)iinfo->lpdib);
	iinfo->lpdib = new_lpdib;
	iinfo->lpdib_size = new_lpdib_size;
	iinfo->lpdib_bits_offs = new_bits_offs;
	
	//iinfo->bitsperpixel = old_bpp;
	//iinfo->palettecolors = pal_entries;
	//iinfo->compression = 0;

	if(has_bitfields) {
		iinfo->bits_red= bitcount[0];
		iinfo->bits_green= bitcount[1];
		iinfo->bits_blue= bitcount[2];
	}
	else if(new_bpp==24) {
		iinfo->bits_red = iinfo->bits_green = iinfo->bits_blue = 8;
	}

	return 1;
}



int pv_bmp_open(HWND hwnd, struct iinfo_struct *iinfo, struct einfo_struct *einfo)
{
	unsigned char fbuf[20];
	DWORD fsize,bytesread;
	HANDLE hfile;
	BOOL b;
	BITMAPCOREHEADER* bc;

	if(iinfo->img_source==0) {  // reading from a file

		hfile=CreateFile(iinfo->img_filename,GENERIC_READ,FILE_SHARE_READ,NULL,
			OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
		if(hfile==INVALID_HANDLE_VALUE) {
			pv_mesg(hwnd,MSG_E,_T("Can't open file"));
			return 0;
		}

		fsize=GetFileSize(hfile,NULL);
		if(fsize==0xffffffff) {
			pv_mesg(hwnd,MSG_E,_T("Can't read file size"));
			CloseHandle(hfile);
			return 0;
		}
		iinfo->file_size = fsize;  // this isn't needed

		// Read 14-byte file header. This is not a part of
		// the DIB in memory.
		ReadFile(hfile,(void*)fbuf,14,&bytesread,NULL);
		if(bytesread<14) {
			pv_mesg(hwnd,MSG_E,_T("Invalid BMP file"));
			CloseHandle(hfile);
			return 0;
		}

		iinfo->lpdib_bits_offs = *(DWORD*)(&fbuf[10]) - 14;
		iinfo->lpdib = (LPBITMAPINFOHEADER)HeapAlloc(einfo->heap,HEAP_ZERO_MEMORY,fsize);
		if(!iinfo->lpdib) {
			pv_mesg(hwnd,MSG_E,_T("Can't allocate memory for BMP"));
			CloseHandle(hfile);
			return 0;
		}
		iinfo->lpdib_size=fsize-14;
		b=ReadFile(hfile,(void*)iinfo->lpdib,iinfo->lpdib_size,&bytesread,NULL);
		CloseHandle(hfile);
		if(!b || (int)bytesread!=iinfo->lpdib_size) {
			pv_mesg(hwnd,MSG_E,_T("Error reading file"));
			HeapFree(einfo->heap,0,(void*)iinfo->lpdib);
			return 0;
		}
		// successfully loaded BMP image
	}
	else if(iinfo->img_source==1) {  // reading from the clipboard
		HGLOBAL hClip;
		unsigned char *lpClip;
		int success = 0;

		OpenClipboard(hwnd);
		hClip=GetClipboardData(CF_DIB);
		if(hClip) {
			// Note that GlobalSize often returns a few more
			// bytes than the DIB uses
			iinfo->lpdib_size = GlobalSize(hClip);
			iinfo->file_size = iinfo->lpdib_size;
			lpClip= (unsigned char*)GlobalLock(hClip);
			if(lpClip) {
				iinfo->lpdib = (LPBITMAPINFOHEADER)HeapAlloc(einfo->heap,HEAP_ZERO_MEMORY,iinfo->lpdib_size);
				if(iinfo->lpdib) {
					CopyMemory((void*)iinfo->lpdib,(void*)lpClip,iinfo->lpdib_size);
					success=1;
				}
				GlobalUnlock(hClip);
			}
		}
		CloseClipboard();
		if(!success) {
			return 0;
		}
	}
	else {
		return 0;
	}

	iinfo->interlaced = 0;
	iinfo->res_flag=0;

	if(iinfo->lpdib->biSize >= 40) {
		if( (iinfo->lpdib->biXPelsPerMeter>0) && (iinfo->lpdib->biYPelsPerMeter>0) )
		{
			iinfo->res_flag=1;
			iinfo->res_x = (int)iinfo->lpdib->biXPelsPerMeter;
			iinfo->res_y = (int)iinfo->lpdib->biYPelsPerMeter;
		}

		switch(iinfo->lpdib->biCompression) {
		case BI_RLE8: iinfo->compression = COMPR_RLE8; break;
		case BI_RLE4: iinfo->compression = COMPR_RLE4; break;
		default: iinfo->compression = COMPR_NONE; break;
		}

		iinfo->bitsperpixel = (int)iinfo->lpdib->biBitCount;

		if(iinfo->lpdib->biBitCount<16) {
			iinfo->colortype = COLORTYPE_PALETTE;

			if(iinfo->lpdib->biClrUsed) {
				iinfo->palettecolors = (int)iinfo->lpdib->biClrUsed;
			}
			else {
				iinfo->palettecolors = 1<<(iinfo->lpdib->biBitCount);
			}
		}
		else {
			iinfo->colortype = COLORTYPE_RGB;
			iinfo->palettecolors = 0;
		}
		iinfo->img_fmt=FMT_BMP;
	}
	else { // old-style bitmap
		bc = (BITMAPCOREHEADER*)iinfo->lpdib;
		iinfo->compression = COMPR_NONE;
		iinfo->bitsperpixel = (int)bc->bcBitCount;

		if(bc->bcBitCount <= 8) {
			iinfo->colortype = COLORTYPE_PALETTE;
			iinfo->palettecolors = 1<<(bc->bcBitCount);
		}
		else {
			iinfo->colortype = COLORTYPE_RGB;
			iinfo->palettecolors = 0;
		}
		iinfo->img_fmt=FMT_BMP;
	}


	if(iinfo->bitsperpixel==24) {
		iinfo->bits_red = iinfo->bits_green = iinfo->bits_blue = 8;
	}

	iinfo->has_bgcolor=0;  // which_bg=0;

	if(!is_restricted_dib(iinfo->lpdib, iinfo->lpdib_size, iinfo->lpdib_bits_offs))
	{
		return fix_dib(hwnd,iinfo,einfo);
	}

	return 1;
}


int pv_bmp_save(HWND hwnd, struct iinfo_struct *iinfo,
				 struct einfo_struct *einfo)
{
	FILE *f;
	BITMAPFILEHEADER h;
	TCHAR fn[MAX_PATH];

	if(iinfo->img_fmt==FMT_BMP && iinfo->img_source==0)
		StringCbCopy(fn,sizeof(fn),iinfo->img_filename);
	else
		StringCbCopy(fn,sizeof(fn),_T("")); 	

	if(!pv_save_as_dialog(hwnd,_T("Save As BMP"),fn,
		_T("BMP image (*.bmp)\0*.bmp\0\0"),_T("bmp")))
	{
		return 0;
	}


	ZeroMemory((void*)&h,sizeof(BITMAPFILEHEADER));
	h.bfType= MAKEWORD('B','M');
	h.bfSize=    sizeof(BITMAPFILEHEADER)+iinfo->lpdib_size;
	h.bfOffBits= sizeof(BITMAPFILEHEADER)+iinfo->lpdib_bits_offs;

	f=_tfopen(fn,_T("wb"));
	if(f) {
		fwrite((void*)&h,1,sizeof(BITMAPFILEHEADER),f);
		fwrite((void*)iinfo->lpdib,1,iinfo->lpdib_size,f);
		fclose(f);
	}
	else {
		pv_mesg(hwnd,MSG_E,_T("Can't save file"));
		return 0;
	}
	return 1;
}
